local M = {}

local helper = require('scenario/scenariohelper')

local min_delivery_distance = 100
local max_delivery_distance = 1500
local pizza_shop_pos = vec3(1053.89319, 1323.19946, 145.01397)

local deliver_pizza_msgs = {"The pizza is getting cold Mario!", "What are you waiting for Mario? Get moving!", "We don't have all day Mario!"}
local thanks_msgs = {"Grazie!", "What took you so long!", "Keep the change!", "Why is my pizza covered in soda!", "Where's the pineapple?!", "Thanks for nothing!"}

-- Difficulty algorithm parameters
local difficulties = 
{
	["easy"] = 
	{
		time_per_dist = 0.06,
		time_added_per_delivery = 15,
		time_lost = 0.25
	},
	["medium"] =
	{
		time_per_dist = 0.055,
		time_added_per_delivery = 10,
		time_lost = 0.5
	},
	["hard"] =
	{
		time_per_dist = 0.05,
		time_added_per_delivery = 5,
		time_lost = 0.75
	}
}

local time_per_dist = 0.06
local time_added_per_delivery = 15
local time_lost = 0.25

local car_released = false
local state = "scenario_started"

local buildings_for_delivery = {}

local the_building_pos = vec3()
local destination_pos = vec3()

local the_wp1_pos = vec3()
local the_wp2_pos = vec3()

local deliveries_made = 0

local distance_to_destination = 0
local current_delivery_distance = 0

local prev_time_elapsed = 0
local prev_time_left = 30
local time_left = 30
local countdown_display_timer = 1
local update_ui_timer = 0
local update_ui_delay = 0.25 -- update UI every 0.25 seconds

local vehicle_selector_active = false
local vehicle_selector_stored_data = nil


local function getState()
	return state
end

local function getDefaultTrafficSize()
	local amount = core_settings_settings.getValue('trafficAmount') -- get amount from gameplay settings
	if amount == 0 then -- use CPU-based value
		amount = getMaxVehicleAmount(20)
	end
	
	return amount
end

local function onScenarioUIReady(scenario_state)
	if scenario_state == "start" then
		if vehicle_selector_active then
			guihooks.trigger('ScenarioChange', vehicle_selector_stored_data)
		else
			guihooks.trigger('MenuHide')
		end
	end
end

local function getDistancesToBuildings(buildings)
	local arr = {}
	
	for key, object in pairs(buildings) do
		local building_pos = vec3(object:getPosition())	
		local distance = (pizza_shop_pos - building_pos):length()
		
		table.insert(arr, {object = object, distance = distance})
	end
	
	return arr
end

local function setPossibleBuildingsForDelivery()
	local objects = getObjectsByClass('TSStatic')

	local datas = getDistancesToBuildings(objects)
	
	for key, data in pairs(datas) do
		local object = data.object
		local distance = data.distance
		
		-- Filtering out bus stops/signs and bridges
		if distance >= min_delivery_distance and distance <= max_delivery_distance and object.shapeName:match("levels/italy/art/shapes/buildings") then	
			if object.shapeName:match("house") 
			or object.shapeName:match("town_bld") 
			then
				table.insert(buildings_for_delivery, object)
			end
		end
	end
end

local function initSettings(difficulty, max_distance, traffic_size)
	local params = difficulties[difficulty]
	
	time_per_dist = params.time_per_dist
	time_added_per_delivery = params.time_added_per_delivery
	time_lost = params.time_lost
	
	max_delivery_distance = max_distance
	
	-- If user set traffic size different from current traffic size
	-- generate new traffic with new size
	if gameplay_traffic.getNumOfTraffic() + 1 ~= traffic_size then
		gameplay_traffic.setupTraffic(traffic_size)
		gameplay_traffic.setTrafficVars({spawnValue = 1.3, spawnDirBias = 0.3})
	end
	
	setPossibleBuildingsForDelivery()
	
	state = "scenario_inited"
end

local function openVehicleSelector()
	-- Code from ge\extensions\flowgraph\nodes\ui\vehicleSelector.lua
	local data = {}
  --data.showDataImmediately = true
  data.introType = 'selectableVehicle'
  data.description = "Its lunch time in Italy and you've got a bunch of hungry customers. Deliver the pizzas as fast as you can without crashing your car before the timer runs out!"
  data.name = "Pizza Run"
  data.portraitText = data.description
  data.portraitImg = {}
  data.portraitImg.start = nil
  data.callObj = 'extensions.scenario_pizza_run_angelo234'
  data.readyHook = data.callObj .. '.vehicleSelectorReady()'
  data.exitHook = data.callObj .. '.vehicleSelectorExited()'
  data.buttonText = "Start"
  data.exitButtonText = nil
  data.vehicle = nil

  vehicle_selector_stored_data = data
  vehicle_selector_active = true
  guihooks.trigger('ChangeState', {state = 'scenario-start', params = {data = data}})
end

-- Called from start screen UI app
local function initSettingsAndSelectVehicle(difficulty, max_distance, traffic_size)
	initSettings(difficulty, max_distance, traffic_size)
	openVehicleSelector()
end

-- close vehicle selector and start scenario
local function closeVehicleSelector()
  guihooks.trigger('MenuHide')
	scenario_scenarios.changeState("running")
	
  vehicle_selector_active = false
end

-- Called from Vehicle Selector UI
local function vehicleSelectorReady()
	closeVehicleSelector()
end

-- Called from Vehicle Selector UI
local function vehicleSelectorExited()
  closeVehicleSelector()
end

-- Called from Vehicle Selector UI
local function onFilteredInputChanged(devName, action, value)

end

local function getClosestRoadPos(building_pos)
	local wp1, wp2, lat_dist_from_wp = map.findClosestRoad(building_pos)
	
	local wp1_pos = vec3(map.getMap().nodes[wp1].pos)
	local wp2_pos = vec3(map.getMap().nodes[wp2].pos)
	
	local xnorm = clamp(building_pos:xnormOnLine(wp1_pos, wp2_pos), 0, 1)
  local wp_dir = wp2_pos - wp1_pos
  local closest_pos_on_road_to_building = xnorm * wp_dir + wp1_pos
	
	local road_width = math.min(map.getMap().nodes[wp1].radius, map.getMap().nodes[wp2].radius)
	
	local offset_vec = math.max(road_width - 1.5, 0) * (building_pos - closest_pos_on_road_to_building):normalized()
	
	return closest_pos_on_road_to_building + offset_vec, wp1, wp2, wp_dir
end

local function getDeliveryTime(dist)
	return (time_per_dist * dist + time_added_per_delivery) - time_lost * deliveries_made
end

local function setDestination(pos, trigger_object)
	local wp1, wp2, wp_dir = nil
	
	destination_pos, wp1, wp2, wp_dir = getClosestRoadPos(pos)
	
	the_wp1_pos = vec3(map.getMap().nodes[wp1].pos)
	the_wp2_pos = vec3(map.getMap().nodes[wp2].pos)
	
	local rot = quatFromDir(wp_dir, vec3(0,0,1))
	
	if trigger_object then
		trigger_object:setPosRot(destination_pos.x, destination_pos.y, destination_pos.z, rot.x, rot.y, rot.z, rot.w)
	end
	
	core_groundMarkers.setFocus(wp1)
	current_delivery_distance = core_groundMarkers.getPathLength()
	
	--core_groundMarkers.setFocus(wp1, math.huge) -- Sets navigation to waypoint and removes ground arrow markers
	--core_groundMarkers.onSerialize() -- Gets rid of floating directional arrows
end

local function getRandomBuilding()
	return buildings_for_delivery[math.random(#buildings_for_delivery)]
end

local function goDeliverPizza()
	prev_time_elapsed = prev_time_left - time_left

	local delivery_point_trigger = scenetree.findObject("delivery_point_trigger")
	
	-- Set destination to random building
	local rand_building = getRandomBuilding()
	--dump(rand_building)
	
	-- getPosition() on C++ side causes runtime error when restarting scenario sometimes
	the_building_pos = vec3(rand_building:getPosition())
	setDestination(the_building_pos, delivery_point_trigger)

	--print("Delivery distance = " .. current_delivery_distance)
	
	local delivery_time = getDeliveryTime(current_delivery_distance)
	time_left = time_left + delivery_time		
	prev_time_left = time_left
	
	local rand_msg = deliver_pizza_msgs[math.random(#deliver_pizza_msgs)]
	
	helper.flashUiMessage(rand_msg .. " The place is " .. string.format("%.2f", current_delivery_distance / 1000.0) .. "km away and you have " .. tostring(math.ceil(time_left)) .. " seconds to get there!", 7)
	
	state = "deliver_pizza"
end

local function goBackToPizzaShop()
	prev_time_elapsed = prev_time_left - time_left

	setDestination(pizza_shop_pos)
	
	--print("Back to shop distance = " .. current_delivery_distance)
	
	local delivery_time = getDeliveryTime(current_delivery_distance)
	time_left = time_left + delivery_time
	prev_time_left = time_left
	
	local rand_msg = thanks_msgs[math.random(#thanks_msgs)]
	
	helper.flashUiMessage(rand_msg .. " You have " .. tostring(math.ceil(time_left)) .. " seconds to go back!", 5)
	
	deliveries_made = deliveries_made + 1
	
	state = "retrieve_pizza"
end

local function onRaceStart()
	goDeliverPizza()
end

local function renderUIText(dt)
	if update_ui_timer >= update_ui_delay then
		local rounded_dist = string.format("%.2f", core_groundMarkers.getPathLength() / 1000.0, 2)
		
		if state == "deliver_pizza" then
			guihooks.message("Deliver the pizza (" .. tostring(rounded_dist) .. "km)", 2 * update_ui_delay, "delivery_status")
		elseif state == "retrieve_pizza" then
			guihooks.message("Go back to pizza shop (" .. tostring(rounded_dist) .. "km)", 2 * update_ui_delay, "delivery_status")
		end
		
		guihooks.message("Time Left: " .. tostring(math.ceil(time_left)) .. "s", 2 * update_ui_delay, "time_left", "timer")
		guihooks.message("Deliveries: " .. tostring(deliveries_made), 2 * update_ui_delay, "num_deliveries", "local_pizza")
	
		update_ui_timer = 0
	end
	
	update_ui_timer = update_ui_timer + dt
end

local function onPreRender(dt, dtSim)
	--[[
	for k,v in pairs(buildings_for_delivery) do
		local pos = vec3(v:getPosition())
	
		local destination_pos = getClosestRoadPos(pos)
		
		debugDrawer:drawCylinder((destination_pos):toPoint3F(), (destination_pos + vec3(0, 0, 20)):toPoint3F(), 1, ColorF(0,1,0,1))
	end
	]]--
	
	local veh = be:getPlayerVehicle(0)
	if not veh or not be:getEnabled() then return end
	
	if state == "scenario_started" or state == "scenario_inited" then
		-- Freeze all cars in case some got unfrozen 
		-- like when selecting new vehicle
		scenario_scenarios.freezeAll(1)
		
		return
	else
		if not car_released then
			scenario_scenarios.freezeAll(0)
			
			car_released = true
		end
	end

	-- Display UI text
	renderUIText(dt)
	
	-- Cylinder to indicate destination
	debugDrawer:drawCylinder((destination_pos + vec3(0, 0, -5)):toPoint3F(), (destination_pos + vec3(0, 0, 75)):toPoint3F(), 1, ColorF(1,0,0,0.25))
	
	if time_left <= 0 then
		-- Game over
		if state == "deliver_pizza" then
			scenario_scenarios.finish({failed = "The pizza became inedible! You delivered a total of " .. tostring(deliveries_made) .. " pizzas."})
		elseif state == "retrieve_pizza" then
			scenario_scenarios.finish({failed = "You didn't make it back on time! You delivered a total of " .. tostring(deliveries_made) .. " pizzas."})
		end
		
		state = "scenario_started"
		
	elseif time_left <= 10 then
		-- When 10 seconds are left: every second display countdown timer
		
		if countdown_display_timer >= 1 then
			helper.flashUiMessage(tostring(math.ceil(time_left)), 0.5, true)	
			
			countdown_display_timer = 0
		end
		countdown_display_timer = countdown_display_timer + dt
		
	else
		countdown_display_timer = 1
		
	end
	
	time_left = time_left - dt
	
	--debugDrawer:drawCylinder((the_building_pos):toPoint3F(), (the_building_pos + vec3(0, 0, 20)):toPoint3F(), 1, ColorF(0,1,0,1))
	--debugDrawer:drawCylinder((the_wp1_pos):toPoint3F(), (the_wp1_pos + vec3(0, 0, 5)):toPoint3F(), 1, ColorF(1,0,0,1))
	--debugDrawer:drawCylinder((the_wp2_pos):toPoint3F(), (the_wp2_pos + vec3(0, 0, 5)):toPoint3F(), 1, ColorF(1,0,0,1))
end

local function onBeamNGTrigger(data)
	local veh = be:getPlayerVehicle(0)
	if not veh or data.subjectName ~= "thePlayer" then return end -- Only care about player on triggers
	
	local stopped = vec3(veh:getVelocity()):length() < 0.5
	
	if stopped then
		if data.triggerName == "delivery_point_trigger" then
			if state == "deliver_pizza" then
				Engine.Audio.playOnce('AudioGui', "event:UI_Checkpoint")
				goBackToPizzaShop()
			end
		elseif data.triggerName == "pizza_shop_trigger" then
			if state == "retrieve_pizza" then		
				Engine.Audio.playOnce('AudioGui', "event:UI_Checkpoint")
				goDeliverPizza()
			end
		end
	end
end

local function onScenarioRestarted(scenario)
  core_groundMarkers.resetAll()
	countdown_display_timer = 1
	prev_time_elapsed = 0
	prev_time_left = 30
	time_left = 30
	deliveries_made = 0
	
	state = "scenario_started"
end

M.getState = getState
M.getDefaultTrafficSize = getDefaultTrafficSize
M.onScenarioLoaded = onScenarioLoaded
M.onScenarioUIReady = onScenarioUIReady
M.initSettings = initSettings
M.initSettingsAndSelectVehicle = initSettingsAndSelectVehicle
M.vehicleSelectorReady = vehicleSelectorReady
M.vehicleSelectorExited = vehicleSelectorExited
M.onFilteredInputChanged = onFilteredInputChanged
M.onRaceStart = onRaceStart
M.onPreRender = onPreRender
M.onBeamNGTrigger = onBeamNGTrigger
M.onScenarioRestarted = onScenarioRestarted

return M